/*
 *   Copyright (c) 2003 David Parsons. All rights reserved.
 *   
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *  3. All advertising materials mentioning features or use of this
 *     software must display the following acknowledgement:
 *     
 *   This product includes software developed by David Parsons
 *   (orc@pell.chi.il.us)
 *
 *  4. My name may not be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *     
 *  THIS SOFTWARE IS PROVIDED BY DAVID PARSONS ``AS IS'' AND ANY
 *  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 *  PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL DAVID
 *  PARSONS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 *  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 *  IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *  THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * transaction handling for files, using LCK. files instead of flock()
 * and friends.
 */

#include "transaction.h"
#include "post.h"

#include <errno.h>
#if defined(__FreeBSD__) || defined(__OpenBSD__)
# include <stdlib.h>		// freebsd has copied malloc.h to stdlib.h
#else
# include <malloc.h>
#endif
#include <fcntl.h>

/*
 *  vpwrollback() abandons the transaction.
 */
int 
vpwrollback(struct vlock *transaction)
{
    if (transaction) {
	unlink(transaction->tempfile);
	if (transaction->fd)
	    fclose(transaction->fd);
	free(transaction->tempfile);
	free(transaction->filename);
	free(transaction);
    }
    return 0;
} /* vpwrollback */


/*
 * vpwcommit() renames the tempfile to the permanent file in a
 * hopefully atomic fashion.
 */
int
vpwcommit(struct vlock *transaction)
{
    int real_err;

    if (transaction) {
	if (transaction->fd) {
	    if (fsync(fileno(transaction->fd)) != 0) {
		real_err = errno;
		vpwrollback(transaction);
		errno = real_err;
		return -1;
	    }
	    if (fclose(transaction->fd) != 0) {
		real_err = errno;
		vpwrollback(transaction);
		errno = real_err;
		return -1;
	    }
	}
	if (rename(transaction->tempfile, transaction->filename) != 0)
	    return -1;
	free(transaction->tempfile);
	free(transaction->filename);
	free(transaction);
	return 0;
    }
    errno = EINVAL;
    return -1;
} /* vpwcommit */


/*
 * vpwlock() locks a file
 */
struct vlock *
vpwlock(char *directory, char *file)
{
    struct vlock *ret = xmalloc(sizeof(struct vlock));
    int saveerr, tick, tfd;

    if (!ret)
	return 0;
    ret->tempfile = concatenate(directory, "/LCK.", file, 0);
    ret->filename = concatenate(directory, "/", file, 0);

    /* The existance of the tempfile is the lock.  Try to open it
     * 6 times, waiting 1 second between tries.
     */
    for (tick=0; tick < 6; sleep(1), tick++) {
	tfd = open(ret->tempfile, O_WRONLY|O_EXCL|O_CREAT, 0644);
	if (tfd != -1)
	    break;
    }
    if (tfd == -1)
	return 0;
    if ( (ret->fd = fdopen(tfd, "w")) == 0) {
	saveerr = errno;
	vpwrollback(ret);
	return 0;
    }
    return ret;
} /* vpwlock */
